#include "S.h"

Date operator+(const Date& a, int diff) {
    Date tmp(a);
    tmp += diff;
    return tmp;
}

bool operator!=(const Date& a, const Date& b) {
    return !(a == b);
}

bool operator>(const Date& a, const Date& b) {
    return b < a;
}

bool operator>=(const Date& a, const Date& b) {
    return !(a < b);
}

bool operator<=(const Date& a, const Date& b) {
    return !(a > b);
}

void prod_to_txt(const std::map<size_t, Production>& prod, const std::string& res_file_name) {

    std::ofstream text(res_file_name);

    text << std::setw(30) << std::left << "STORAGE ID" << std::setw(30) << std::left
        << "PRODUCT CODE" << std::setw(30) << std::left << "PRODUCT NAME" << std::setw(30)
        << std::left << "ARRIVAL DATE" << std::setw(30) << std::left << "SHELF LIFE IN DAYS"
        << std::setw(30) << std::left << "AMOUNT OF PRODUCTS" << std::setw(30) << std::left << "UNIT PRICE(c.u.)" << std::endl;

    if (!text.is_open()) {
        throw std::runtime_error("Error. The file is not opened!");
    }
    for (const auto& p : prod) {
        text << p.second;
    }
    text.close();
}

void Storage::new_prod_insert(Production& tmp_prod) {
    std::cout << "Please input the data of the new product\n";
    std::cin >> tmp_prod;
}

void Storage::load_in() {
    std::ifstream in_f(bin_f_name, std::ios::binary | std::ios::out | std::ios::in);
    Production tmp_prod;

    while (tmp_prod.binary_input(in_f)) {
        f_amount++;
        product.emplace(f_amount, tmp_prod);
    }
    in_f.close();
}

Storage::Storage(const std::string& bin_f, const std::string& txt_to_out) : bin_f_name(bin_f), txt_f_name(txt_to_out) {
    load_in();
}

Storage::~Storage() {
    product.clear();
    bin_f_name.clear();
    txt_f_name.clear();
}

void Storage::create_and_write() {
    std::cout << std::endl << "Creating new bin file" << std::endl;
    if (!product.empty()) {
        f_amount = 0;

        // ЧТОБЫ ОЧИСТИТЬ ФАЙЛ, НУЖЕН ФЛАГ trunc
        // ДА И ЗАЧЕМ ВООБЩЕ ПРЕДВАРИТЕЛЬНО ОТКРЫВАТЬ ФАЙЛ ДЛЯ ОЧИСТКИ?
        // МОЖНО СРАЗУ ОТКРЫТЬ ЕГО С ФЛАГОМ trunc И ФАЙЛ БУДЕТ ПУСТОЙ

    }
    std::ofstream new_b_file(bin_f_name, std::ios::binary | std::ios::out | std::ios::trunc);
    if (!new_b_file.is_open()) {
        throw std::runtime_error("Error. The file is not opened!");

    }
    else {
        std::cout << "The file has opened succesfully" << std::endl;
        char choice = 'q';
        while (choice != 'n') {
            std::cout << "Want to add new data?" << std::endl << "y - yes" << std::endl << "n - no" << std::endl;
            std::cin >> choice;
            if (choice == 'n') {
                break;
            }
            else if (choice == 'y') {
                f_amount++;
                Production newProd;
                new_prod_insert(newProd);
                newProd.binary_output(new_b_file);
                product.emplace(f_amount, newProd);
            }
        }
        new_b_file.close();
    }

}

void Storage::update_file() {
    std::ofstream curr_file(bin_f_name, std::ios::binary | std::ios::out);
    if (!curr_file.is_open()) {
        throw std::out_of_range("Error. The file is not opened!");

    }
    for (auto& p : product) {
        p.second.binary_output(curr_file);
    }
    curr_file.close();
}

void Storage::add_product() {
    std::ofstream curr_File(bin_f_name, std::ios::binary | std::ios::app);
    if (!curr_File.is_open()) {
        throw std::runtime_error("Error.The file is not opened!");
    }
    else {
        std::cout << std::endl << "Adding new product..." << std::endl;
        Production newprod;
        f_amount++;
        new_prod_insert(newprod);
        product.insert(std::make_pair(f_amount, newprod));
        newprod.binary_output(curr_File);
        curr_File.close();
    }
}

void Storage::delete_product(int prod_id) {
    if (product.count(prod_id)) {
        throw std::out_of_range("There is no such id in the journal!");
    }
    product.erase(prod_id);
    f_amount--;
    update_file();
}


void Storage::change_product(int prod_id) {
    auto update_iterator = product.find(prod_id);
    new_prod_insert(update_iterator->second);
    update_file();
}

void Storage::search_storage_id(int id) noexcept {
    std::map<size_t, Production> result_map;
    for (const auto& [key, value] : product) {
        if (value.storage_id != id) continue;
        result_map.emplace(key, value);
    }
    if (result_map.empty()) {
        std::cout << std::endl << "There are no products in this storage" << std::endl;
    }
    else {
        std::cout << std::endl << "Products in this storage are(check it in the result.txt)" << std::endl;
        prod_to_txt(result_map, "result.txt");
    }
}

void Storage::search_product_code(int code) noexcept {
    std::map<size_t, Production> result_map;
    for (const auto& [key, value] : product) {
        if (value.product_code != code) continue;
        result_map.emplace(key, value);
    }
    if (result_map.empty()) {
        std::cout << std::endl << "There are no products with this code" << std::endl;
    }
    else {
        std::cout << std::endl << "Products with this code are(check it in the result.txt)" << std::endl;
        prod_to_txt(result_map, "result.txt");

    }
}

void Storage::search_arrival_date(const Date& arr_date) noexcept {
    std::map<size_t, Production> result_map;
    for (const auto& [key, value] : product) {
        if (value.arrival_date != arr_date) continue;
        result_map.emplace(key, value);
    }
    if (result_map.empty()) {
        std::cout << std::endl << "There are no products with this arrivcal date" << std::endl;
    }
    else {
        std::cout << std::endl << "Products whith this arrival(check it in the result.txt)" << std::endl;
        prod_to_txt(result_map, "result.txt");
    }
}

void Storage::search_overdue_products(const Date& curr_date) noexcept {
    std::map<size_t, Production> result_map;
    for (const auto& [key, value] : product) {
        if (value.arrival_date + value.shelf_life <= curr_date) continue;
        result_map.emplace(key, value);
    }
    if (result_map.empty()) {
        std::cout << std::endl << "There are no expired products" << std::endl;
    }
    else {
        std::cout << std::endl << "Expired products are(check it in the result.txt)" << std::endl;
        prod_to_txt(result_map, "result.txt");

    }
}

void Storage::search_not_overdue_products(const Date& curr_date) noexcept {
    std::map<size_t, Production> result_map;
    std::map<size_t, Production>::iterator current_iter = product.begin();
    for (const auto& [key, value] : product) {
        if (value.arrival_date + value.shelf_life > curr_date) continue;
        result_map.emplace(key, value);
    }
    if (result_map.empty()) {
        std::cout << std::endl << "There are no not-expired products" << std::endl;
    }
    else {
        std::cout << std::endl << "Not-expired products are(check it in the result.txt)" << std::endl;
        prod_to_txt(result_map, "result.txt");
    }
}

void Storage::load_out_txt() {

    std::ofstream text(txt_f_name);

    text << std::setw(30) << std::left << "STORAGE ID" << std::setw(30) << std::left
        << "PRODUCT CODE" << std::setw(30) << std::left << "PRODUCT NAME" << std::setw(30)
        << std::left << "ARRIVAL DATE" << std::setw(30) << std::left << "SHELF LIFE IN DAYS"
        << std::setw(30) << std::left << "AMOUNT OF PRODUCTS" << std::setw(30) << std::left << "UNIT PRICE(c.u.)" << std::endl;

    double all_summ = 0.;
    for (auto& p : product) {
        text << p.second;
        all_summ += (p.second.amount_of_products * p.second.unit_price);
    }

    text << std::endl << "--------------------------------------------------------------" << std::endl;
    text << "The amount of produts in the journal:" << f_amount << std::endl << "Total price of all the products is: " << all_summ;
    text.close();
}
