#pragma once
#include <map>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <iterator>
#include <iomanip>
#include <algorithm>
#include <string>
#include <cstring>

struct Date {
    int day;
    int month;
    int year;

    bool operator==(const Date& to_compare) const {
        return (day == to_compare.day) && (month == to_compare.month) && (year == to_compare.year);
    }

    bool operator<(const Date& to_compare) const {
        if (year != to_compare.year) {
            return year < to_compare.year;
        }
        if (month != to_compare.month) {
            return month < to_compare.month;
        }
        if (day != to_compare.day) {
            return day < to_compare.day;
        }
        return false;
    }

    int day_in_month(int year, int month) const {
        static constexpr int kFebruary = 1;
        static const std::vector<int> days{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
        if (month == kFebruary && ((year % 4 == 0) && (year % 100 != 0) || (year % 400 == 0))) {
            return 29;
        }
        return days[month];
    }

    Date& operator+=(int diff) {
        while (day + diff > day_in_month(year, month)) {
            diff -= day_in_month(year, month) - day;
            day = 0;
            if (month == 12) {
                year++;
                month = 0;
            }
            month++;
        }
        day += diff;
        return *this;
    }

/*
    Date operator+(int day) const {
        Date ret(*this);
        while (ret.day + day > day_in_month(ret.year, ret.month)) {
            day = day - (day_in_month(ret.year, ret.month) - ret.day);
            ret.day = 0;
            if (ret.month == 12) {
                ret.year++;
                ret.month = 0;
            }
            ret.month++;
        }
        ret.day += day;

        return ret;
    }
*/


    friend std::ostream& operator << (std::ostream& out, const Date& p_date) {
        out << p_date.day << "." << p_date.month << "." << p_date.year;
        return out;
    };

    friend std::istream& operator >> (std::istream& in, Date& p_date) {
        std::cout << "day = ";
        in >> p_date.day;
        std::cout << "month = ";
        in >> p_date.month;
        std::cout << "year = ";
        in >> p_date.year;
        return in;
    }
};

Date operator+(const Date& a, int diff);

bool operator!=(const Date& a, const Date& b);

bool operator>(const Date& a, const Date& b);

bool operator>=(const Date& a, const Date& b);

bool operator<=(const Date& a, const Date& b);

struct Production {
    int storage_id;
    int product_code;
    std::string product_name;
    Date arrival_date;
    int shelf_life;
    int amount_of_products;
    double unit_price;

    bool operator==(const Production& prod_to_comp) {
        return storage_id == prod_to_comp.storage_id &&
            product_code == prod_to_comp.product_code &&
            product_name == prod_to_comp.product_name &&
            arrival_date == prod_to_comp.arrival_date &&
            shelf_life == prod_to_comp.shelf_life &&
            amount_of_products == prod_to_comp.amount_of_products &&
            // std::abs(unit_price - prod_to_comp.unit_price) < 1e-6;
            unit_price == prod_to_comp.unit_price;
    }

    Production() = default;
    Production(const Production& elem) = default;
    Production& operator=(const Production& elem) = default;


    friend std::ostream& operator<<(std::ostream& out, const Production& P) {
        out << std::setw(40) << std::left << P.storage_id << std::setw(45) << std::left << P.product_code
            << std::setw(50) << std::left << P.product_name << P.arrival_date
            << std::setw(50) << std::right << P.shelf_life << std::setw(40) << std::right << P.amount_of_products
            << std::setw(50) << std::right << P.unit_price << std::endl;

        return out;
    }

    friend std::istream& operator >>(std::istream& in, Production& P) {
        std::cout << "Storage id = ";
        in >> P.storage_id;
        std::cout << "Product code = ";
        in >> P.product_code;
        std::cout << "Product name = ";
        in >> P.product_name;
        std::cout << "Arrival date:" << std::endl;
        in >> P.arrival_date;
        std::cout << "Shelf life in days = ";
        in >> P.shelf_life;
        std::cout << "Amount of products = ";
        in >> P.amount_of_products;
        std::cout << "Unit price = ";
        in >> P.unit_price;
        return in;

    }

    struct Proto {
        int storage_id;
        int product_code;
        int product_name_length;
        Date arrival_date;
        int shelf_life;
        int amount_of_products;
        double unit_price;
    };

    Proto to_proto() const {
        Proto proto;
        proto.amount_of_products = amount_of_products;
        proto.arrival_date = arrival_date;
        proto.product_code = product_code;
        proto.product_name_length = product_name.length();
        proto.shelf_life = shelf_life;
        proto.storage_id = storage_id;
        proto.unit_price = unit_price;
        return proto;
    }

    void from_proto(const Proto& proto) {
        amount_of_products = proto.amount_of_products;
        arrival_date = proto.arrival_date;
        product_code = proto.product_code;
        shelf_life = proto.shelf_life;
        storage_id = proto.storage_id;
        unit_price = proto.unit_price;
    }

    size_t to_buffer(char* buffer) const {
        Proto proto = to_proto();
        size_t length = sizeof(proto) + product_name.length();
        std::memcpy(buffer, &proto, sizeof(proto));
        std::memcpy(buffer + sizeof(proto), product_name.c_str(), product_name.length());
        return length;
    }

    bool from_buffer(void* buffer) {
        Proto proto;
        std::memcpy(&proto, buffer, sizeof(proto));
        auto v = std::make_unique<char[]>(proto.product_name_length);
        std::memcpy(v.get(), static_cast<char*>(buffer) + sizeof(proto), proto.product_name_length);
        product_name = v.get();
        return true;
    }

    bool binary_input(std::ifstream& in) {
        Proto proto;
        bool ans = static_cast<bool>(in.read(reinterpret_cast<char*>(&proto), sizeof(proto)));
        if (!ans) return false;
        auto buffer = std::make_unique<char[]>(proto.product_name_length + sizeof(proto));
        in.read(buffer.get() + sizeof(proto), proto.product_name_length);
        std::memcpy(buffer.get(), &proto, sizeof(proto));
        from_buffer(buffer.get());
        return true;
    }

    void binary_output(std::ofstream& out) const {
        auto buffer = std::make_unique<char[]>(sizeof(Proto) + product_name.length());
        size_t sz = to_buffer(buffer.get());
        out.write(buffer.get(), sz);
    }
};

class Storage {
private:
    std::map<size_t, Production> product;
    std::string bin_f_name;
    std::string txt_f_name;
    size_t f_amount;

    static void new_prod_insert(Production& tmp_prod);
    void load_in();

public:
    Storage(const std::string& bin_f, const std::string& txt_to_out);
    ~Storage();

    void create_and_write();
    void print_data()  noexcept;
    void print_product(const std::pair<size_t, Production>& prod);
    void update_file();

    void add_product();
    void delete_product(int prod_id);
    void change_product(int prod_id);

    void search_storage_id(int id) noexcept;
    void search_product_code(int code) noexcept;
    void search_arrival_date(const Date& arr_date) noexcept;
    void search_overdue_products(const Date& curr_date) noexcept;
    void search_not_overdue_products(const Date& curr_date) noexcept;

    void load_out_txt();
};
